import click
import numpy as np
from numpy.random import default_rng
from tqdm import tqdm

from radiosim.jet import create_jet
from radiosim.mojave import create_mojave
from radiosim.survey import create_survey
from radiosim.utils import (
    _save_mojave_bundle,
    add_noise,
    adjust_outpath,
    create_grid,
    save_sky_distribution_bundle,
)

__all__ = ["create_sky_distribution", "simulate_sky_distributions"]


def simulate_sky_distributions(conf):
    if conf["mode"] == "mojave":
        seed = conf["seed"]
        if seed == "none":
            rng = default_rng()
        elif isinstance(seed, int):
            rng = default_rng(seed)
        else:
            raise TypeError('seed has to be int or "none"')
    else:
        rng = None
    for opt in ["train", "valid", "test"]:
        create_sky_distribution(
            conf=conf,
            opt=opt,
            rng=rng,
        )


def create_sky_distribution(conf, opt: str, rng=None) -> None:
    for _ in tqdm(range(conf["bundles_" + opt])):
        path = adjust_outpath(conf["outpath"], f"/data_{conf['mode']}_" + opt)
        grid = create_grid(conf["img_size"], conf["bundle_size"])
        if conf["mode"] == "jet":
            sky, target = create_jet(grid, conf)
        elif conf["mode"] == "survey":
            sky, target = create_survey(grid, conf)
        elif conf["mode"] == "mojave":
            data, data_name = create_mojave(conf, rng)
            sky_bundle = data[0].copy()
            if conf["noise"]:
                sky_bundle = np.squeeze(
                    add_noise(np.expand_dims(sky_bundle, axis=1), conf["noise_level"]),
                    axis=1,
                )
                # for img in data:
                #     img -= img.min()
                #     img /= img.max()
            _save_mojave_bundle(path, data=[sky_bundle, *data[1:]], data_name=data_name)
            continue
        else:
            click.echo(
                "Given mode not found. Choose 'survey', "
                "'jet' or 'mojave' in config file"
            )

        sky_bundle = sky.copy()
        target_bundle = target.copy()
        if conf["noise"] and conf["noise_level"] > 0:
            sky_bundle = add_noise(sky_bundle, conf["noise_level"])
            for img in sky_bundle:
                img -= img.min()
                img /= img.max()

        save_sky_distribution_bundle(path, sky_bundle, target_bundle)



#!/usr/bin/env python3
import datetime
import os
import tomllib
from pathlib import Path

import radiosim

pyproject_path = Path(__file__).parent.parent / "pyproject.toml"
pyproject = tomllib.loads(pyproject_path.read_text())

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinxcontrib.bibtex",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "IPython.sphinxext.ipython_console_highlighting",
    "matplotlib.sphinxext.plot_directive",
    "myst_parser",
    "numpydoc",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_tippy",
    "sphinx_togglebutton",
]

numpydoc_show_class_members = False
numpydoc_class_members_toctree = False

myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
]

myst_url_schemes = {
    "http": None,
    "https": None,
    "wiki": "https://en.wikipedia.org/wiki/{{path}}#{{fragment}}",
    "doi": "https://doi.org/{{path}}",
    "gh-issue": {
        "url": "https://github.com/radionets-project/radiosim/issue/{{path}}#{{fragment}}",
        "title": "Issue #{{path}}",
        "classes": ["github"],
    },
}

bibtex_bibfiles = ["references.bib"]
bibtex_encoding = "utf8"

copybutton_exclude = ".linenos, .gp"
copybutton_selector = "div:not(.no-copybutton) > div.highlight > pre"
copybutton_image_svg = """
<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 640 640"><!--!Font Awesome Free v7.0.0 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free Copyright 2025 Fonticons, Inc.-->
    <path d="M480 400L288 400C279.2 400 272 392.8 272 384L272 128C272 119.2 279.2 112 288 112L421.5 112C425.7 112 429.8 113.7 432.8 116.7L491.3 175.2C494.3 178.2 496 182.3 496 186.5L496 384C496 392.8 488.8 400 480 400zM288 448L480 448C515.3 448 544 419.3 544 384L544 186.5C544 169.5 537.3 153.2 525.3 141.2L466.7 82.7C454.7 70.7 438.5 64 421.5 64L288 64C252.7 64 224 92.7 224 128L224 384C224 419.3 252.7 448 288 448zM160 192C124.7 192 96 220.7 96 256L96 512C96 547.3 124.7 576 160 576L352 576C387.3 576 416 547.3 416 512L416 496L368 496L368 512C368 520.8 360.8 528 352 528L160 528C151.2 528 144 520.8 144 512L144 256C144 247.2 151.2 240 160 240L176 240L176 192L160 192z"/>
</svg>
"""  # noqa: E501

templates_path = ["_templates"]
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "changes",
    "*.log",
]

source_suffix = {".rst": "restructuredtext", ".md": "markdown"}
master_doc = "index"


project = pyproject["project"]["name"]
author = pyproject["project"]["authors"][0]["name"]
copyright = "{}.  Last updated {}".format(
    author, datetime.datetime.now().strftime("%d %b %Y %H:%M")
)
python_requires = pyproject["project"]["requires-python"]

# make some variables available to each page
rst_epilog = f"""
.. |python_requires| replace:: {python_requires}
"""


version = radiosim.__version__
# The full version, including alpha/beta/rc tags.
release = version


# -- Version switcher -----------------------------------------------------

# Define the json_url for our version switcher.
json_url = "https://radiosim.readthedocs.io/en/latest/_static/switcher.json"

# Define the version we use for matching in the version switcher.,
version_match = os.getenv("READTHEDOCS_VERSION")
# If READTHEDOCS_VERSION doesn't exist, we're not on RTD
# If it is an integer, we're in a PR build and the version isn't correct.
if not version_match or version_match.isdigit():
    # For local development, infer the version to match from the package.
    version_match = "latest" if "dev" in release or "rc" in release else release

    # We want to keep the relative reference when on a pull request or locally
    json_url = "_static/switcher.json"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_file_suffix = ".html"

html_css_files = [
    "css/radiosim.css",
    "css/sphinx-design.css",
    "css/tippy.css",
]

html_favicon = "_static/favicon/favicon.ico"

html_theme_options = {
    "github_url": "https://github.com/radionets-project/radiosim",
    "header_links_before_dropdown": 5,
    "navbar_start": ["navbar-logo", "version-switcher"],
    "switcher": {
        "version_match": version_match,
        "json_url": json_url,
    },
    "navigation_with_keys": False,
    "show_version_warning_banner": True,
    "icon_links_label": "Quick Links",
    "icon_links": [
        {
            "name": "Radionets Project",
            "url": "https://github.com/radionets-project",
            "type": "url",
            "icon": "https://avatars.githubusercontent.com/u/77392854?s=200&v=4",  # noqa: E501
        },
    ],
    "logo": {
        "text": "radiosim",
    },
    "announcement": """
        <p>radiosim is not stable yet, so expect large and rapid
        changes to structure and functionality as we explore various
        design choices before the 1.0 release.</p>
    """,
}

html_title = f"{project}: Radio Galaxy Simulations in Python"
htmlhelp_basename = project + "docs"


# Configuration for intersphinx
intersphinx_mapping = {
    "astropy": ("https://docs.astropy.org/en/stable", None),
    "ipywidgets": ("https://ipywidgets.readthedocs.io/en/stable", None),
    "joblib": ("https://joblib.readthedocs.io/en/stable", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numba": ("https://numba.readthedocs.io/en/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "pytest": ("https://docs.pytest.org/en/stable", None),
    "python": ("https://docs.python.org/3", None),
    "rich": ("https://rich.readthedocs.io/en/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "torch": ("https://pytorch.org/docs/stable/", None),
}


suppress_warnings = [
    "intersphinx.external",
]



import os
import re
import sys
from pathlib import Path

import click
import cv2
import h5py
import numpy as np
import toml
from astropy.convolution import Gaussian2DKernel
from numpy.typing import ArrayLike
from scipy import signal

__all__ = [
    "add_noise",
    "adjust_outpath",
    "cart2pol",
    "check_outpath",
    "create_grid",
    "load_data",
    "pol2cart",
    "read_config",
    "relativistic_boosting",
    "save_sky_distribution_bundle",
    "zoom_on_source",
    "zoom_out",
]


def create_grid(pixel, bundle_size):
    """
    Creates a square 2d grid.

    Parameters
    ----------
    pixel: int
        number of pixel in x and y
    bundle_size: int
        number of images in each bundle

    Returns
    -------
    grid: ndarray
        2d grid with 1e-10 pixels, X meshgrid, Y meshgrid
    """
    x = np.linspace(0, pixel - 1, num=pixel)
    y = np.linspace(0, pixel - 1, num=pixel)
    X, Y = np.meshgrid(x, y)
    grid = np.array([np.zeros(X.shape) + 1e-10, X, Y])
    grid = np.repeat(
        grid[None, :, :, :],
        bundle_size,
        axis=0,
    )
    return grid


def relativistic_boosting(theta, beta):
    """
    Calculate relativistic boosting factor for a jet.

    Parameters
    ----------
    theta: float
        angle of the jet in relation to the observer
    beta: float
        velocity of the jet components

    Returns
    -------
    boost_app: float
        boosting factor for the approaching jet
    boost_rec: float
        boosting factor for the receding jet
    """
    gamma = 1 / np.sqrt(1 - beta**2)  # Lorentz factor
    mu = np.cos(theta)

    boost_app = 1 / (gamma * (1 - beta * mu))
    boost_rec = 1 / (gamma * (1 + beta * mu))
    return boost_app, boost_rec


def zoom_on_source(img, comp=None, max_amp=0.01):
    """
    Zoom on source to cut out irrelevant area. Shape will stay equal.

    Parameters
    ----------
    img: 2D array
        Image of the sky used to zoom on
    comp: 3D array (n, (img))
        Images of the components, same zooming as on img
    max_amp: float
        Maximal amplitude which will be at the edge of the image

    Returns
    -------
    zoomed_img: ndarray
        Image after zooming
    zoom_factor: float
        Zooming factor
    """
    # find farest outside column or row with amplitude > max_amp
    mask = img > max_amp
    mask_flip = np.flip(mask)

    idx_left = np.argmax(np.sum(mask, axis=0) > 0)
    idx_right = np.argmax(np.sum(mask_flip, axis=0) > 0)
    idx_bottom = np.argmax(np.sum(mask, axis=1) > 0)
    idx_top = np.argmax(np.sum(mask_flip, axis=1) > 0)
    # print(idx_left, idx_right, idx_bottom, idx_top)
    idx = np.min([idx_left, idx_right, idx_bottom, idx_top])
    size = img.shape[0]
    zoom_factor = size / (size - 2 * idx)

    # crop the source
    cropped_img = img[idx : size - idx, idx : size - idx]
    zoomed_img = cv2.resize(
        cropped_img, dsize=(size, size), interpolation=cv2.INTER_LINEAR
    )

    if comp is not None:
        cropped_comp = comp[:, idx : size - idx, idx : size - idx]
        zoomed_comp = np.empty_like(comp)
        for i, component in enumerate(cropped_comp):
            zoomed_comp[i] = cv2.resize(
                component, dsize=(size, size), interpolation=cv2.INTER_LINEAR
            )
        return zoomed_img, zoomed_comp, zoom_factor

    return zoomed_img, zoom_factor


def zoom_out(img, comp=None, pad_value=0):
    """
    Zoom out of an image. Padding edges with zeros.

    Parameters
    ----------
    img: 2D array
        Image of the sky
    comp: 3D array (n, (img))
        Images of the components
    pad_value: int
        Number of pixels to pad around the source

    Returns
    -------
    zoomed_img: ndarray
        Image after zooming
    zoomed_comp: ndarray
        Componets after zooming
    """
    if not isinstance(pad_value, int):
        pad_value = np.int64(pad_value)
    size = img.shape[0]
    img = cv2.resize(
        np.pad(img, pad_value), dsize=(size, size), interpolation=cv2.INTER_LINEAR
    )

    if comp is not None:
        for component in comp:
            component = cv2.resize(
                np.pad(component, pad_value),
                dsize=(size, size),
                interpolation=cv2.INTER_LINEAR,
            )
        return img, comp

    return img


def check_outpath(outpath, quiet=False):
    """
    Check if outpath exists. Check for existing source_bundle files.
    Ask to overwrite or reuse existing files.

    Parameters
    ----------
    outpath : str
        path to out directory
    quiet : bool
        activate quiet mode, overwrite existing files

    Returns
    -------
    sim_sources : bool
        flag to enable/disable source simulation routine
    """
    path = Path(outpath)
    exists = path.exists()
    if exists is True:
        source = {p for p in path.rglob("*data*.h5") if p.is_file()}
        if source:
            click.echo("Found existing source simulations!")
            if quiet or click.confirm(
                "Do you really want to overwrite existing files?", abort=False
            ):
                click.echo("Overwriting existing source simulations!")
                [p.unlink() for p in source]
                sim_sources = True
                return sim_sources
            else:
                click.echo("Keeping existing source simulations!")
                sim_sources = False
                sys.exit()
        else:
            sim_sources = True
    else:
        Path(path).mkdir(parents=True, exist_ok=False)
        sim_sources = True
    return sim_sources


def read_config(config):
    """
    Unpacking of the config file to print the config parameters

    Parameters
    ----------
    config: toml-file
        toml configuration file with all parameters
    Returns
    -------
    sim_comf: dictionary
        unpacked configurations
    """
    sim_conf = {}
    sim_conf["seed"] = config["general"]["seed"]
    sim_conf["mode"] = config["general"]["mode"]
    sim_conf["threads"] = config["general"]["threads"]
    sim_conf["outpath"] = config["paths"]["outpath"]
    sim_conf["training_type"] = config["jet"]["training_type"]
    sim_conf["num_jet_components"] = config["jet"]["num_jet_components"]
    sim_conf["scaling"] = config["jet"]["scaling"]
    sim_conf["num_sources"] = config["survey"]["num_sources"]
    sim_conf["class_distribution"] = config["survey"]["class_distribution"]
    sim_conf["scale_sources"] = config["survey"]["scale_sources"]
    sim_conf["class_ratio"] = config["mojave"]["class_ratio"]
    sim_conf["bundles_train"] = config["image_options"]["bundles_train"]
    sim_conf["bundles_valid"] = config["image_options"]["bundles_valid"]
    sim_conf["bundles_test"] = config["image_options"]["bundles_test"]
    sim_conf["bundle_size"] = config["image_options"]["bundle_size"]
    sim_conf["img_size"] = config["image_options"]["img_size"]
    sim_conf["noise"] = config["image_options"]["noise"]
    sim_conf["noise_level"] = config["image_options"]["noise_level"]

    return sim_conf


def add_noise(image, noise_level):
    """
    Used for adding noise.

    Parameters
    ----------
    image: 4darray
        bundle of images of shape [n, 1, size, size]
    noise_level: int
        maximum intensity of noise in percent

    Returns
    -------
    image_noised: 4darray
        bundle of noised images
    """

    def noise_small(kernel, mean=0, std=1):
        """
        Create the noise of different kernel sizes
        """
        max_noise = np.random.uniform(0, 1, img_shape[0])
        noise = (
            np.random.normal(mean, std, size=img_shape) * max_noise[:, None, None, None]
        )
        g_kernel = Gaussian2DKernel(kernel / 2).array[None, None, :]
        return signal.fftconvolve(noise, g_kernel, mode="same")

    def call_noise(kernels, strengths):
        """
        Loop through lists and normalize
        """
        noise_out = np.zeros(shape=img_shape)
        for kernel, strength in zip(kernels, strengths):
            if kernel == 1:
                noise = np.random.normal(size=img_shape)
            else:
                noise = noise_small(kernel)
            noise /= np.abs(noise).max() / strength
            noise_out += noise
        return noise_out

    img_shape = image.shape
    kernels = [1, img_shape[-1] / 32, img_shape[-1] / 8]
    strengths = [0.2, 0.3, 0.5]  # have to add up to 1

    noise = call_noise(kernels, strengths)

    noise_level = (
        np.random.uniform(*noise_level, size=image.shape[0])[:, None, None, None] / 100
    )

    noise /= noise.max(axis=(1, 2, 3))[:, None, None, None]
    noise *= image.max(axis=(1, 2, 3))[:, None, None, None] * noise_level
    image_noised = noise + image

    return image_noised


def adjust_outpath(path, option, form="h5"):
    """
    Add number to out path when filename already exists.

    Parameters
    ----------
    path: str
        path to save directory
    option: str
        additional keyword to add to path
    form: str
        file extension

    Returns
    -------
    out: str
        adjusted path
    """
    counter = 0
    filename = str(path) + (option + "_{}." + form)
    while os.path.isfile(filename.format(counter)):
        counter += 1
    out = filename.format(counter)
    return out


def save_sky_distribution_bundle(path, x, y, name_x="x", name_y="y"):
    """
    Write images created in analysis to h5 file.

    Parameters
    ----------
    path: str
        path to save file
    x: ndarray
        image of the full jet, sum over all components
    y: ndarray
        images of components or list, depends on train_type
    name_x: str
        name of the x-data
    name_y: str
        name of the y-data
    """
    with h5py.File(path, "w") as hf:
        hf.create_dataset(name_x, data=x)
        hf.create_dataset(name_y, data=y)
        hf.close()


def _save_mojave_bundle(path: str, data: tuple, data_name: tuple) -> None:
    """
    Write MOJAVE simulations created in analysis to h5 file.

    Parameters
    ----------
    path: str
        path to save file
    data : tuple
        data to store, e.g. galaxies, classes, coordinates, u2026
    data_name : tuple
        name of the data columns
    """
    assert len(data) == len(data_name)
    with h5py.File(path, "w") as hf:
        for dat, name in zip(data, data_name):
            hf.create_dataset(name, data=dat)
        hf.close()


def _gen_vlba_obs_position(rng, size: int) -> tuple[float, float]:
    ra = rng.uniform(0, 360, size=size)
    dec = rng.uniform(-30, 87, size=size)

    return ra, dec


def _gen_date(rng, start_date: str, size: int) -> ArrayLike:
    from datetime import datetime

    from h5py import string_dtype

    # define fixed string length for h5py
    utf8_type = string_dtype("utf-8", 10)

    start_date = np.datetime64(start_date)
    now = np.datetime64(datetime.now().date())
    delta = now - start_date
    dates = start_date + rng.integers(delta.astype(int), size=size)

    return dates.astype(utf8_type)


def cart2pol(x: float, y: float):
    """
    Transforms cartesian to polar coordinates.

    Parameters
    ----------
    x: float
        x-coordinate
    y: float
        y-coordinate

    Returns
    -------
    r: float
        radius, euclidean distance between (0,0) and (x,y)
    phi: float
        angle in radian
    """
    r = np.sqrt(x**2 + y**2)
    phi = np.arctan2(y, x)

    return (r, phi)


def pol2cart(r: float, phi: float):
    """
    Transforms polar to cartesian coordinates.

     Parameters
    ----------
    r: float
        radius, euclidean distance between (0,0) and (x,y)
    phi: float
        angle in radian

    Returns
    -------
    x: float
        x-coordinate
    y: float
        y-coordinate
    """
    x = r * np.cos(phi)
    y = r * np.sin(phi)

    return (x, y)


def load_data(conf_path, data_type="train", key="x"):
    config = toml.load(conf_path)
    path = Path(config["paths"]["outpath"])
    bundle_paths = np.array([x for x in path.iterdir()])
    paths = [
        path for path in bundle_paths if re.findall("data*" + data_type, path.name)
    ]
    data = []
    for path_test in paths:
        df = h5py.File(path_test, "r")
        data.extend(df[key])

    return data



import click
import numpy as np
import toml

from radiosim.simulations import simulate_sky_distributions
from radiosim.utils import check_outpath, read_config


@click.command()
@click.argument("configuration_path", type=click.Path(exists=True, dir_okay=False))
@click.option(
    "--mode",
    type=click.Choice(
        [
            "simulate",
            "overview",
        ],
        case_sensitive=False,
    ),
    default="simulate",
)
def main(configuration_path, mode) -> None:
    config = toml.load(configuration_path)
    conf = read_config(config)
    print(conf, "\n")

    if conf["seed"] != "none":
        click.echo(f"Using numpy random seed {conf['seed']}. \n")
        np.random.seed(conf["seed"])

    outpath = config["paths"]["outpath"]
    sim_sources = check_outpath(
        outpath,
        quiet=config["general"]["quiet"],
    )

    if mode == "simulate" and sim_sources:
        simulate_sky_distributions(conf)


if __name__ == "__main__":
    main()



import numpy as np
from astropy.convolution import Gaussian2DKernel
from astropy.nddata.utils import Cutout2D
from scipy.stats import norm, skewnorm

__all__ = [
    "gauss",
    "skewed_gauss",
    "twodgaussian",
]


def gauss(params: list, size: int) -> np.ndarray:
    """
    Returns a 2d gaussian distribution. Creates a gaussian in the center with twice
    the image size and then cuts out the correct position.
    Pro: Easy to understand and maintain
    Con: Slower than twodgaussian; maybe faster, if the cutout could be removed

    Parameters
    ----------
    params: list
        [amplitude, center_x, center_y, width_x, width_y, rot]
        (rot (rotation) in radian)
    shape: int
        length of the image

    Returns
    -------
    gaussian_scaled: 2darray
        gaussian distribution in two dimensions
    """
    gaussian_2D_kernel = Gaussian2DKernel(
        x_stddev=params[3],
        y_stddev=params[4],
        theta=params[5],
        x_size=size * 2,
        y_size=size * 2,
    )
    gaussian_cutout = Cutout2D(
        data=gaussian_2D_kernel,
        position=(size * 1.5 - params[1], size * 1.5 - params[2]),
        size=(size, size),
    ).data
    gaussian_scaled = gaussian_cutout / gaussian_cutout.max() * params[0]

    return gaussian_scaled


def twodgaussian(params: list, size: int) -> np.ndarray:
    r"""
    Returns a 2d gaussian function of the form:

    .. math::

       x' &= \cos(rot) * x - \sin(rot) * y \\
       y' &= \sin(rot) * x + \cos(rot) * y \\
       g  &= a * \exp(-(((x - \mathtt{center\_x})/ \mathtt{width\_x})^2 \\
          &\phantom{=} + ((y - \mathtt{center\_y}) / \mathtt{width\_y})^2 ) / 2 )

    - Pro: Faster than gauss
    - Con: More difficult to understand

    Parameters
    ----------
    params: list
        [amplitude, center_x, center_y, width_x, width_y, rot]
        (rot (rotation) in radian)
    size: int
        length of the image

    Returns
    -------
    rotgauss: array_like
        gaussian distribution in two dimensions

    Notes
    -----
    Short version of ``twodgaussian``:
    https://github.com/keflavich/gaussfitter/blob/0891cd3605ab5ba000c2d5e1300dd21c15eee1fd/gaussfitter/gaussfitter.py#L75
    """
    amplitude = params[0]
    center_y, center_x = params[1], params[2]
    width_y, width_x = params[3], params[4]
    rot = params[5]
    rcen_x = center_x * np.cos(rot) - center_y * np.sin(rot)
    rcen_y = center_x * np.sin(rot) + center_y * np.cos(rot)

    def rotgauss(x, y):
        xp = x * np.cos(rot) - y * np.sin(rot)
        yp = x * np.sin(rot) + y * np.cos(rot)
        g = amplitude * np.exp(
            -(((rcen_x - xp) / width_x) ** 2 + ((rcen_y - yp) / width_y) ** 2) / 2.0
        )
        return g

    return rotgauss(*np.indices((size, size)))


def skewed_gauss(
    size: int,
    x: float,
    y: float,
    amp: float,
    width: float,
    length: float,
    a: float,
) -> np.ndarray:
    """
    Generate a skewed 2d normal distribution.

    Parameters
    ----------
    size: int
        length of the square image
    x: float
        x coordinate of the center
    y: float
        y coordinate of the center
    amp: float
        maximal amplitude of the distribution
    width: float
        width of the distribution (perpendicular to the skewed function)
    length: float
        length of the distribution
    a: float
        skewness parameter

    Returns
    -------
    skew_gauss: array_like
        two dimensional skewed gaussian distribution
    """

    def skewfunc(x: float, **args) -> np.ndarray:
        func = skewnorm.pdf(x, **args)
        func /= func.max()
        return func

    vals = np.arange(size)
    normal = norm.pdf(vals, loc=y, scale=width).reshape(-1, 1)
    normal /= normal.max()

    skew = skewfunc(vals, a=a, loc=x, scale=length)

    skew_gauss = normal * skew
    skew_gauss *= amp
    skew_gauss[np.isclose(skew_gauss, 0)] = 0

    return skew_gauss



from pathlib import Path

import click
import h5py
import numpy as np
import toml
from tqdm import tqdm

from radiosim.ppdisks import generate_proto_set


@click.command()
@click.argument(
    "configuration_path",
    type=click.Path(exists=True, dir_okay=False),
)
def main(configuration_path) -> None:
    config = toml.load(configuration_path)
    rng = np.random.default_rng(seed=config["metadata"]["seed"])

    outpath = Path(config["dataset"]["outpath"])

    if not outpath.is_dir():
        outpath.mkdir()

    file_prefix = config["dataset"]["file_prefix"]

    metadata_params = config["metadata"]
    device = config["simulation"]["device"]
    verbose = config["simulation"]["verbose"]

    batch_num = config["dataset"]["batches"]
    batch_size = config["dataset"]["batch_size"]
    batches = {
        "train": int(batch_num * config["dataset"]["batches_train_ratio"]),
        "valid": int(batch_num * config["dataset"]["batches_valid_ratio"]),
        "test": int(batch_num * config["dataset"]["batches_test_ratio"]),
    }

    # resolution and pointings taken from FITS headers of DSHARP dataset
    # https://almascience.eso.org/almadata/lp/DSHARP/
    resolution = 0.00299999999999988
    pointings = [
        (252.1901583333, -14.27663888889),
        (242.8806334167, -18.64062026556),
        (246.6002916667, -24.27041666667),
        (269.088663625, -21.9562676325),
        (236.6862890833, -34.51002105694),
        (246.912515125, -23.97199338639),
        (246.6875916667, -24.38563138889),
        (239.1667003167, -22.02789646667),
        (246.484, -24.34672222222),
        (246.5986583334, -24.72051666667),
        (246.578625, -24.47213611111),
        (246.9142916667, -24.65431666667),
        (239.653748625, -22.95433409167),
        (252.313727625, -14.36918004583),
        (239.0382866667, -37.93514649278),
        (236.3035304583, -34.29194894083),
        (239.1762312125, -37.82110333889),
        (239.8185583333, -41.95297045306),
        (242.2576419583, -39.08689564417),
        (240.1854510899, -41.92538824062),
    ]

    for batch_type, num in batches.items():
        for i in tqdm(np.arange(num), desc=f"Generating {batch_type}"):
            protos, params = generate_proto_set(
                img_size=metadata_params["img_size"],
                size=batch_size,
                alpha_range=metadata_params["alpha_range"],
                ratio_range=metadata_params["ratio_range"],
                size_ratio_range=metadata_params["size_ratio_range"],
                device=device,
                seed=rng.integers(low=0, high=2**32 - 1),
                verbose=verbose,
            )
            metadata = []
            for j in np.arange(batch_size):
                pointing = rng.choice(pointings)
                metadata.append(
                    dict(
                        index=j,
                        cell_size=resolution,
                        src_ra=pointing[0],
                        src_dec=pointing[1],
                    )
                )

            with h5py.File(outpath / f"{file_prefix}_{batch_type}_{i}.h5", "w") as hf:
                hf.create_dataset("y", data=protos)
                hf.create_dataset(
                    "metadata", data=str(metadata), dtype=h5py.string_dtype()
                )
                hf.create_dataset("params", data=str(params), dtype=h5py.string_dtype())
                hf.close()


if __name__ == "__main__":
    main()



import numpy as np
from joblib import Parallel, delayed
from scipy.stats import expon, skewnorm
from skimage.transform import rotate, swirl

from radiosim.gauss import skewed_gauss, twodgaussian
from radiosim.utils import _gen_date, _gen_vlba_obs_position

__all__ = [
    "add_swirl",
    "create_mojave",
    "gen_compact",
    "gen_jet",
    "gen_one_jet",
    "gen_shocks",
    "gen_two_jet",
]


def create_mojave(conf, rng):
    """
    Create MOJAVE like sources.

    Parameters:
    -----------
    conf :
        loaded conf file

    Returns:
    --------
    glx : array_like
        generated sources
    """

    threads = conf["threads"]
    if threads == "none":
        threads = 1
    elif not isinstance(threads, int) or threads <= 0:
        raise ValueError("threads has to be int >0 or 'none'")

    size = conf["img_size"]

    # calculate amount of each class to generate per bundle
    bundle_size = conf["bundle_size"]
    ratio = np.array(conf["class_ratio"])

    # amount = np.array([n_comact, n_one_jet, n_two_jet]])
    amount = np.array((ratio / ratio.sum()) * bundle_size).astype(int)
    while amount.sum() < bundle_size:
        amount[amount.argmax()] += 1

    jets = []
    compact = Parallel(n_jobs=threads)(
        delayed(gen_compact)(rng=child, size=size) for child in rng.spawn(amount[0])
    )
    one_jet = Parallel(n_jobs=threads)(
        delayed(gen_one_jet)(rng=child, size=size) for child in rng.spawn(amount[1])
    )
    two_jet = Parallel(n_jobs=threads)(
        delayed(gen_two_jet)(rng=child, size=size) for child in rng.spawn(amount[2])
    )
    jets = np.array([*compact, *one_jet, *two_jet])

    glx_class = []
    for glx_type, amt in zip([0, 1, 2], amount):
        glx_class.extend([glx_type] * amt)
    glx_class = np.array(glx_class)

    shuffler = np.arange(bundle_size)
    rng.shuffle(shuffler)

    ra, dec = _gen_vlba_obs_position(rng=rng, size=bundle_size)
    obs_dates = _gen_date(rng=rng, start_date="1995-01-01", size=bundle_size)

    data = [jets[shuffler], glx_class[shuffler], ra[shuffler], dec[shuffler], obs_dates]
    data_name = ["galaxies", "galaxy_classes", "RA", "DEC", "date"]

    return data, data_name


def gen_jet(
    size: int, amplitude: float, width: float, length: float, a: float
) -> np.ndarray:
    """
    Generate jet from skewed 2d normal distributiuon.

    Parameters
    ----------
    size : int
        length of the square image
    amplitude : float
        maximal amplitude of the distribution
    width : float
        width of the distribution (perpendicular to the skewed function)
    length : float
        length of the distribution
    a : float
        skewness parameter

    Returns
    -------
    jet : array_like
        jet for source
    """

    jet = skewed_gauss(
        size=size,
        x=size / 2,
        y=size / 2,
        amp=amplitude,
        width=width,
        length=length,
        a=a,
    )

    jet[np.isclose(jet, 0)] = 0

    return jet


def gen_shocks(
    jet: np.ndarray,
    shock_comp: int,
    length: float,
    width: float,
    sx: float,
    rng: np.random.Generator,
) -> tuple[np.ndarray, tuple, int]:
    """
    Generate equally spaced shocks in jet.

    Parameters
    ----------
    jet : array_like
        generated jet without shock
    shock_comp : int
        max amount of shocks
    length : float
        length of jet
    width : float
        width of jet
    sx : float
        sx component of main gaussian
    rng : :class:`~numpy.random.Generator`
        numpy random generator

    Returns
    -------
    jet : array_like
        jet with shock components
    printout_information : tuple
        information about each component
    skipped : int
        amount of skipped shock components
    """
    ps_orientation = []
    ps_amp = []
    ps_x = []
    ps_y = []
    ps_sx = []
    ps_sy = []
    ps_rot = []
    ps_dist = []
    ps_a = []

    mask = np.copy(jet)
    mask[np.isclose(mask, 0)] = 0
    mask[mask > 0] = 1
    mask = mask.astype(bool)

    size = len(jet)

    s_length = 2 * length - sx
    s_dist = s_length / shock_comp
    s_dist0 = s_dist
    skipped = 0
    for _ in range(shock_comp):
        # skip randomly
        if rng.uniform(0, 1) > 2 / 3:
            skipped += 1
            continue

        s_x = size / 2 + s_dist
        s_y = size / 2
        s_amp = jet[int(s_y), int(s_x)] * rng.uniform(0.03, 0.6)

        s_sx = rng.uniform(*np.sort([2, width]))
        s_sy = rng.uniform(*np.sort([s_sx, length / shock_comp]))

        s_rot = 0

        s_a = rng.uniform(5, 9)

        jet += skewed_gauss(size, int(s_x), int(s_y), s_amp, s_sx, s_sy, s_a)

        ps_amp.append(s_amp)
        ps_x.append(int(s_x))
        ps_y.append(int(s_y))
        ps_sx.append(s_sx)
        ps_sy.append(s_sy)
        ps_rot.append(s_rot)
        ps_dist.append(s_dist)
        ps_a.append(s_a)
        ps_orientation.append("r")

        s_dist += s_dist0

    return (
        jet,
        (ps_orientation, ps_amp, ps_x, ps_y, ps_sx, ps_sy, ps_rot, ps_dist, ps_a),
        skipped,
    )


def add_swirl(
    jet: np.ndarray,
    rng: np.random.Generator,
    first_jet_params: tuple | None = None,
) -> tuple[np.ndarray, tuple]:
    """
    Add swirl distortion to the jet.

    Parameters
    ----------
    jet : array_like
        generated jet
    rng : :class:`~numpy.random.Generator`
        numpy random generator
    first_jet_params : tuple | None, default: None
        Use when generate two jet sources.
        Tuple of parameters to generate similar swirl for second jet.
        Contains the returned parameters from the first jet.

    Returns
    -------
    swirled_jet : array_like
        swirl distorted input jet
    parameters : tuple
        parameters of the aplied swirl distortion
    """
    size = len(jet)
    if not first_jet_params:
        strength = rng.normal(loc=0, scale=0.2)
        radius = rng.uniform(50, 200)
        center = np.array([size / 2, size / 2])
    else:
        deviation = 1 + rng.normal(loc=0, scale=0.1, size=2)
        strength = first_jet_params[0] * deviation[0]
        radius = first_jet_params[1] * deviation[1]
        center = first_jet_params[2]

    swirled_jet = swirl(jet, center=center, strength=strength, radius=radius)

    return swirled_jet, (strength, radius, center)


def gen_two_jet(
    rng: int,
    size: int = 1024,
    printout: bool = False,
) -> np.ndarray:
    """
    Generate a two jet galaxy.

    Parameters
    ----------
    rng : :class:`~numpy.random.Generator` | int
    numpy random generator or seed to generate random number generator
    size : int, default: 1024
        size of the galaxy to be genrated
    printout : bool, default: False
        print out log information

    Returns
    -------
    glx : array_like
        simulated two jet source
    """
    if isinstance(rng, int):
        rng = np.random.default_rng(seed=rng)

    amp_params = (1.715e-4, 1.985e-2)
    width_params = (4.37, 22.71, 16.85)
    length_params = (8.62, 69.60, 54.73)
    left_deviation = 1 + rng.normal(
        loc=0, scale=0.1, size=5
    )  # amp, length, width, strength, radius

    amp = expon.rvs(*amp_params, random_state=rng)

    r_width = skewnorm.rvs(*width_params, random_state=rng) / 10  # 10
    r_length = skewnorm.rvs(*length_params, random_state=rng) / 4  # 10
    r_jet_amp = amp * rng.power(0.7)

    l_width = r_width * left_deviation[2]
    l_length = r_length * left_deviation[1]
    l_jet_amp = r_jet_amp * left_deviation[0]

    dimensions = np.sort([r_width / 2, r_length / 5, l_width / 2, l_length / 5])

    sx = rng.uniform(dimensions.min(), dimensions.max()) * 1.5  # too large
    sy = rng.uniform(dimensions.min(), dimensions.max()) * 1.5

    # redraw if gaussian is too elliptical
    while np.any([sx / sy < 1 / 2, sx / sy > 2]):
        sy = rng.uniform(dimensions.min(), dimensions.max()) * 1.5
    rot = rng.uniform(0, 2 * np.pi)

    a = 4

    r_jet = gen_jet(size, r_jet_amp, r_width, r_length, a)
    l_jet = gen_jet(size, l_jet_amp, l_width, l_length, a)

    shock_comp_r = int(np.round(rng.power(0.5), decimals=1) * 10)
    shock_comp_l = int(np.round(rng.power(0.5), decimals=1) * 10)

    if shock_comp_r > 0:
        r_jet, r_printout, r_skipped = gen_shocks(
            r_jet, shock_comp_r, r_length, r_width, sx, rng
        )

    if shock_comp_l > 0:
        l_jet, l_printout, l_skipped = gen_shocks(
            l_jet, shock_comp_l, l_length, l_width, sx, rng
        )

    r_jet, r_swirl_comp = add_swirl(r_jet, rng=rng)
    l_jet, l_swirl_comp = add_swirl(l_jet, rng=rng, first_jet_params=r_swirl_comp)

    glx = r_jet + np.flip(l_jet)
    x = size / 2
    y = size / 2

    gauss = twodgaussian([amp, size / 2, size / 2, sx, sy, rot], size)
    glx += gauss

    angle = rng.uniform(0, 360)

    glx = rotate(glx, angle=angle)

    glx[np.isclose(glx, 0)] = 0

    if printout:
        print("left Jet:")
        print(
            f"width = {l_width:.3f}, length = {l_length:.3f}, \
                jet_amp = {l_jet_amp:.3f}, a = {a:.3f}"
        )
        print("right Jet:")
        print(
            f"width = {r_width:.3f}, length = {r_length:.3f}, \
                jet_amp = {r_jet_amp:.3f}, a = {a:.3f}"
        )
        print("Source:")
        print(
            f"amp = {amp:.3f}, x = {x:.0f}, y = {y:.0f}, sx = {sx:.3f}, \
                sy = {sy:.3f}, rot = {rot:.3f}"
        )
        if shock_comp_r > 0:
            print(
                f"Right Shock, {shock_comp_r} Components, \
                    {r_skipped} skipped:"
            )
            for _, amp, x, y, sx, sy, rot, dist, _ in zip(*r_printout):
                print(
                    f"amp = {amp:.6f}, x = {x:.0f}, y = {y:.0f}, \
                        sx = {sx:.3f}, sy = {sy:.3f}, rot = {rot:.3f}, \
                        dist = {dist:.3f}"
                )
        if shock_comp_l > 0:
            print(
                f"Left Shock, {shock_comp_l} Components, \
                    {l_skipped} skipped:"
            )
            for _, amp, x, y, sx, sy, rot, dist, _ in zip(*l_printout):
                print(
                    f"amp = {amp:.6f}, x = {x:.0f}, y = {y:.0f}, \
                        sx = {sx:.3f}, sy = {sy:.3f}, rot = {rot:.3f}, \
                        dist = {dist:.3f}"
                )
        print("Swirl:")
        print(
            f"Right: strength = {r_swirl_comp[0]:.3f}, \
                radius = {r_swirl_comp[1]:.3f}, center = {r_swirl_comp[2]}"
        )
        print(
            f"Left: strength = {l_swirl_comp[0]:.3f}, \
                radius = {l_swirl_comp[1]:.3f}, center = {l_swirl_comp[2]}"
        )

    return glx


def gen_one_jet(rng: int, size: int = 1024, printout: bool = False) -> np.ndarray:
    """
    Generate a one jet galaxy.

    Parameters
    ----------
    rng : :class:`~numpy.random.Generator` | int
        numpy generator or seed to generate random number generator
    size : int, default: 1024
        size of the galaxy to be genrated
    printout : bool, default: False
        print out debug information

    Returns
    -------
    glx : ndarray
        simulated one jet source
    """
    if isinstance(rng, int):
        rng = np.random.default_rng(seed=rng)

    amp_params = (1.715e-4, 1.985e-2)
    width_params = (4.37, 22.71, 16.85)
    length_params = (8.62, 69.60, 54.73)

    amp = expon.rvs(*amp_params, random_state=rng)

    width = skewnorm.rvs(*width_params, random_state=rng) / 12  # 10
    length = skewnorm.rvs(*length_params, random_state=rng) / 6.5  # 10
    jet_amp = amp * rng.power(0.7)

    dimensions = np.sort([width / 2, length / 5])

    sx = rng.uniform(*dimensions) * 1.5  # too large
    sy = rng.uniform(*dimensions) * 1.5

    # redraw if gaussian is too elliptical
    while np.any([sx / sy < 1 / 2, sx / sy > 2]):
        sy = rng.uniform(dimensions.min(), dimensions.max()) * 1.5
    rot = rng.uniform(0, 2 * np.pi)

    a = 4

    jet = gen_jet(size, jet_amp, width, length, a)

    shock_comp = int(np.round(rng.power(0.5), decimals=1) * 10)

    if shock_comp > 0:
        jet, shock_printout, skipped = gen_shocks(
            jet, shock_comp, length, width, sx, rng
        )

    jet, swirl_comp = add_swirl(jet, rng=rng)

    x = size / 2
    y = size / 2

    gauss = twodgaussian([amp, size / 2, size / 2, sx, sy, rot], size)
    glx = jet + gauss

    angle = rng.uniform(0, 360)

    glx = rotate(glx, angle=angle)

    glx[np.isclose(glx, 0)] = 0

    if printout:
        print("Skewed dist:")
        print(
            f"width = {width:.3f}, length = {length:.3f}, \
                jet_amp = {jet_amp:.3f}, a = {a:.3f}"
        )
        print("Source:")
        print(
            f"amp = {amp:.3f}, x = {x:.0f}, y = {y:.0f}, sx = {sx:.3f}, \
                sy = {sy:.3f}, rot = {rot:.3f}"
        )
        if shock_comp > 0:
            print(f"{shock_comp} shock components, {skipped} skipped:")
            for _, amp, x, y, sx, sy, rot, dist, _ in zip(shock_printout):
                print(
                    f"amp = {amp:.6f}, x = {x:.0f}, y = {y:.0f}, \
                        sx = {sx:.3f}, sy = {sy:.3f}, rot = {rot:.3f}, \
                        dist = {dist:.3f}"
                )
        print("Swirl:")
        print(
            f"strength = {swirl_comp[0]:.3f}, \
                radius = {swirl_comp[1]:.3f}, center = {swirl_comp[2]}"
        )

    return glx


def gen_compact(rng: int, size: int = 1024, printout: bool = False) -> np.ndarray:
    """Generate a compact galaxy.

    Parameters
    ----------
    rng : :class:`~numpy.random.Generator` | int
        numpy generator or seed to generate random number generator.
    size : int, default: 1024
        size of the galaxy to be genrated
    printout : bool, default: False
        print out debug information.

    Returns
    -------
    glx : ndarray
        Simulated compact source.
    """
    if isinstance(rng, int):
        rng = np.random.default_rng(seed=rng)

    amp_params = (5.54e-05, 0.02)
    width_params = (8.41, 3.04, 3.46)
    length_params = (6.04, 3.02, 2.59)

    ncomps = rng.integers(1, 4)  # draw amount of components
    outflow = rng.uniform() > 0.5  # draw if outflow of the jet is generated

    amp = expon.rvs(*amp_params, random_state=rng)

    jet_amp = amp * rng.normal(2 / 3, 0.05)

    y = np.arange(size)

    width = skewnorm.rvs(*width_params, random_state=rng) * 0.8
    length = skewnorm.rvs(*length_params, random_state=rng) * 1

    if outflow:
        a = 4
        glx = gen_jet(size, jet_amp, width, length, a)
        glx, swirl_comp = add_swirl(glx, rng=rng)
    else:
        glx = np.zeros((size, size))

    dimensions = np.sort([width / 2, length / 4])
    dimensions[0] *= 1.2
    dimensions = np.sort(dimensions)

    i = 0
    comp_amp = []
    comp_x = []
    comp_y = []
    comp_sx = []
    comp_sy = []
    comp_rot = []
    vertical = True
    while i < ncomps:
        deviation = 1 + rng.normal(loc=0, scale=0.2, size=2)
        if i != 0:
            x = rng.uniform(-5, 5)
            y = rng.uniform(-5, 5)
        else:
            x = 0
            y = 0
        if i == 0:
            sx = rng.uniform(*dimensions) * 1.7  # too large
            j = 0
            while sx > 15:
                if j > 10:
                    sx /= 5
                    break
                sx = rng.uniform(*dimensions) * 1.7
                j += 1

            sy = rng.uniform(*dimensions) * 1.7

            j = 0
            while sy > 15:
                if j > 10:
                    sy /= 5
                    break
                sy = rng.uniform(*dimensions) * 1.7
                j += 1

            j = 0
            while np.any([sx / sy < 0.33, sx / sy > 3]) or np.all(
                [sx / sy > 0.75, sx / sy < 1.33]
            ):
                if j > 10:
                    sy = sx * rng.choice([rng.uniform(0.4, 0.8), rng.uniform(1.2, 2.5)])
                    if sy > 15:
                        sy /= 5
                    break
                sy = rng.uniform(*dimensions) * 1.7
                k = 0
                while sy > 15:
                    if k > 10:
                        sy /= 5
                        break
                    sy = rng.uniform(*dimensions) * 1.7
                    k += 1
                j += 1
        else:
            sx = comp_sx[-1] * deviation[0]
            sy = comp_sy[-1] * deviation[1]

        if i == 0:
            vertical = sx > sy

        try:
            rot = rng.uniform(comp_rot[-1] - 0.1, comp_rot[-1] + 0.1)
        except IndexError:
            if vertical:
                rot = rng.uniform(-np.pi / 4, np.pi / 4)
            else:
                rot = rng.uniform(-np.pi / 4 + np.pi / 2, np.pi / 4 + np.pi / 2)

        c_amp = amp * rng.uniform(0.5, 1.2)

        gauss = twodgaussian([c_amp, size / 2 - x, size / 2 - y, sx, sy, rot], size)
        glx += gauss

        comp_amp.append(c_amp)
        comp_x.append(size / 2 - x)
        comp_y.append(size / 2 - y)
        comp_sx.append(sx)
        comp_sy.append(sy)
        comp_rot.append(rot)

        i += 1

    angle = rng.uniform(0, 360)
    glx = rotate(glx, angle=angle)

    shift = size // 2 - np.argwhere(glx == glx.max())[0]

    glx = np.roll(glx, shift=shift, axis=[0, 1])

    glx[np.isclose(glx, 0)] = 0

    glx /= glx.max()
    glx *= amp

    if printout:
        if outflow:
            print("Skewed dist:")
            print(
                f"width = {width:.3f}, "
                "length = {length:.3f}, "
                "jet_amp = {jet_amp:.3f}, a = {a:.3f}"
            )
            print("Swirl:")
            print(
                f"strength = {swirl_comp[0]:.3f},"
                "radius = {swirl_comp[1]:.3f}, "
                "center = {swirl_comp[2]}"
            )
        print("Sources:")
        for cx, cy, csx, csy, cr in zip(comp_x, comp_y, comp_sx, comp_sy, comp_rot):
            print(
                f"amp = {amp:.3f},"
                f"x = {cx:.0f}, "
                f"y = {cy:.0f}, "
                f"sx = {csx:.3f}, "
                f"sy = {csy:.3f}, "
                f"rot = {cr:.3f}"
            )

    return glx



def test_generate_proto_set():
    import h5py
    import numpy as np
    import torch

    from radiosim.ppdisks import generate_proto_set

    sizes = [1, 2]
    img_sizes = [128, 512]
    size_ratio_ranges = [(0.1, 1), (0.1, 0.5), (0.5, 1)]
    seed = 42

    images = dict()

    for img_size in img_sizes:
        images[str(img_size)] = []

    for size in sizes:
        for img_size in img_sizes:
            for size_ratio_range in size_ratio_ranges:
                protos, params = generate_proto_set(
                    img_size=img_size,
                    size=size,
                    size_ratio_range=size_ratio_range,
                    seed=seed,
                )
                images[str(img_size)].extend(protos)

    for img_size in img_sizes:
        with h5py.File(f"./tests/ppdisks_{img_size}.h5", "r") as hf:
            assert torch.allclose(
                torch.from_numpy(hf["y"][()]),
                torch.from_numpy(np.array(images[str(img_size)])),
            )


def test_generate_proto_no_seed():
    from radiosim.ppdisks import create_proto

    img_sizes = [128, 512]
    alphas = [0.0, 180.0]
    size_ratios = [0.1, 1.0]
    ratios = [3, 15]

    for img_size in img_sizes:
        for alpha in alphas:
            for ratio in ratios:
                for size_ratio in size_ratios:
                    create_proto(
                        img_size=img_size,
                        alpha=alpha,
                        ratio=ratio,
                        size_ratio=size_ratio,
                        seed=None,
                    )



# Code adapted from Andreas Maisinger (TU Dortmund University)

import numpy as np

try:
    import torch
except ImportError as err:  # pragma: no cover
    raise ModuleNotFoundError(
        "You need torch to use the ppdisks module of radiosim!"
        'Use uv pip install "radiosim[torch]" to install the necessary packages!'
    ) from err

from astropy.convolution import Gaussian2DKernel
from torch.fft import fft2, ifft2
from torchvision.transforms.functional import resize
from tqdm.autonotebook import tqdm

__all__ = ["create_proto", "generate_proto_set"]


def generate_proto_set(
    img_size: int,
    size: int,
    alpha_range: tuple = (0, 180),
    ratio_range: tuple = (3, 15),
    size_ratio_range: tuple = (0.1, 1),
    device: str = "cpu",
    seed: int = 1337,
    verbose: bool = False,
):
    """
    Generates a Set of simulated protoplanetary disk images

    Parameters
    ----------
    img_size: int
        The size of the image (height and width)
    size: int
        The amount of images to simulate
    alpha_range: tuple of float, optional
        The range of values of the inclination of the
        disk in degrees. Default: (0, 180)
    ratio_range: tuple of float, optional
        The range of values of the ratio the minor axis should
        be of the major axis :math:`a / b`. Default: (3, 15)
    size_ratio_range: tuple of float, optional
        The range of values of the ratio of the image
        size the disk should take up. Default: (0.1, 1)
    device: str, optional
        The name of the device to run the simulations
        on (with torch). Default: 'cpu'
    seed: int, optional
        The seed for the random generator. Default: 1337
    verbose: bool, optional
        Whether to show a progress bar or not. Default: False

    Returns
    -------
    protos: list of torch.tensor
        List of the simulated images

    params: dict
        The used simulation parameters

    """

    np.random.seed(seed)

    bundles = tqdm(range(size)) if verbose else range(size)

    outputs = [
        create_proto(
            img_size,
            alpha=np.random.randint(*alpha_range),
            ratio=np.random.randint(*ratio_range) / np.random.randint(*ratio_range),
            size_ratio=np.random.uniform(*size_ratio_range),
            device=device,
        )
        for i in bundles
    ]

    protos = [out.cpu().numpy() for out in outputs]
    params = dict(
        seed=seed,
        N=size,
        img_size=img_size,
        alpha_range=alpha_range,
        ratio_range=ratio_range,
        size_ratio_range=size_ratio_range,
    )

    return protos, params


def create_proto(
    img_size: int,
    alpha: float,
    ratio: float,
    size_ratio: float,
    device: str = "cpu",
    seed: int = None,
):
    """
    Generates a simulated protoplanetary disk image

    Parameters
    ----------
    img_size: int
        The size of the image (height and width)
    alpha: float
        The inclination angle of the disk in degrees
    ratio: float
        The ratio the minor axis should be of the major axis a / b
    size_ratio: float
        The ratio of the image size the disk should take up
    device: str, optional
        The name of the device to run the simulations on (with torch)
    seed: int, optional
        The seed for the random generator. If set to ``None``, the seed is ignored.

    Returns
    -------
    proto: torch.tensor
        The simulated image of the protoplanetary disk

    """

    if seed is not None:
        np.random.seed(seed)

    fix_size = 900
    center = (fix_size // 2, fix_size // 2)
    m, n = _get_m_n()
    core, d_core = _sim_core(fix_size, alpha, center, ratio, device=device)
    sc = np.random.randint(5, 10)
    core = _smooth(sc, sc, core, device=device) * _get_scale(d_core, m, n)
    num_circ = int(np.random.normal(3, 0.5) // 1)

    els_full = torch.zeros((900, 900), device=device)
    ds = []
    a_max = d_core

    for i in range(num_circ):
        eli, di, maxlength = _sim_ring(
            fix_size, alpha, center, ratio, m, n, (i + 1) * 30, device=device
        )
        a_max = np.max([maxlength, a_max])
        els_full += eli
        ds.append(di)

    proto = core + els_full
    proto = torch.abs(proto)

    img_size2 = int(fix_size // 2)

    a_max = np.min([img_size2, a_max])

    proto_snip = proto[
        int(img_size2 - a_max) : int(img_size2 + a_max),
        int(img_size2 - a_max) : int(img_size2 + a_max),
    ]

    if proto_snip.max() < 1e-10:
        return create_proto(img_size, alpha, ratio, size_ratio)

    new_size = int(size_ratio * img_size)
    new_size = new_size if new_size % 2 == 0 else new_size - 1

    proto_snip = proto_snip[None, :, :]

    proto_snip = resize(proto_snip, (new_size, new_size))
    proto_snip = proto_snip[0]

    proto_snip_size2 = int(proto_snip.shape[0] // 2)

    img_size2 = int(img_size // 2)

    proto = torch.ones((img_size, img_size), device=device) * 1e-10
    proto[
        img_size2 - proto_snip_size2 : img_size2 + proto_snip_size2,
        img_size2 - proto_snip_size2 : img_size2 + proto_snip_size2,
    ] += proto_snip

    del proto_snip
    return proto.cpu()


def _create_ellipse(
    img_size: int,
    a: float,
    b: float,
    center: tuple,
    alpha: float,
    device: str = "cpu",
):
    """
    Generates an ellipse

    Parameters
    ----------
    img_size: int
        The size of the image (height and width)
    a: float
        The major axis of the ellipse
    b: float
        The minor axis of the ellipse
    center: tuple of float
        The center of the ellipse
    alpha: float
        The rotation angle of the ellipse
    device: str, optional
        The name of the device to run the simulations on (with torch)

    Returns
    -------
    ellipse: torch.tensor
        The image of the ellipse
    average_distance: float
        The average distance from the center of the ellipse
    """

    x_0 = center[0]
    y_0 = center[1]

    xx, yy = torch.from_numpy(np.mgrid[:img_size, :img_size]).to(device)
    term1 = (
        ((xx - x_0) * np.cos(np.deg2rad(alpha)))
        + ((yy - y_0) * np.sin(np.deg2rad(alpha)))
    ) ** 2
    term2 = (
        ((xx - x_0) * np.sin(np.deg2rad(alpha)))
        - ((yy - y_0) * np.cos(np.deg2rad(alpha)))
    ) ** 2
    ellipse = ((term1 / a**2) + (term2 / b**2)) <= 1
    avarage_distance = (a + b) / 2
    return ellipse, avarage_distance


def _substract(
    ellipse1: torch.tensor, ellipse2: torch.tensor, distance1: float, distance2: float
):
    """
    Substracts two ellipses from eachother

    Parameters
    ----------
    ellipse1: torch.tensor
    The first ellipse

    ellipse2: torch.tensor
    The second ellipse

    distance1: float
    The average distance from the center of the first ellipse

    distance1: float
    The average distance from the center of the second ellipse

    Returns
    -------
    subtracted: torch.tensor
    The result of the subtraction

    average_distance: float
    The average distance from the center of the ellipse resulting from the subtraction

    """

    subtracted = ellipse1 ^ ellipse2
    average_distance = (distance1 + distance2) / 2
    return subtracted, average_distance


def _sim_core(
    img_size: int, alpha: float, center: tuple, ratio: float, device: str = "cpu"
):
    """
    Simulates the core of the protoplanetary disk

    img_size: int
    The size of the image (height and width)

    alpha: float
    The rotation angle of the core

    center: tuple of float
    The center of the core

    ratio: float
    The ratio the minor axis should be of the major axis a / b

    device: str, optional
    The name of the device to run the simulations on (with torch)

    Returns
    -------
    core: torch.tensor
    The image of the core

    d_core: float
    The average distance from the center of the core

    """

    a = np.float64(np.random.randint(3, 15))
    core, d_core = _create_ellipse(img_size, a, a / ratio, center, alpha, device=device)
    return core, d_core


def _sim_ring(
    img_size: int,
    alpha: float,
    center: tuple,
    ratio: float,
    m: float,
    n: float,
    initial_a: float = 0,
    device: str = "cpu",
):
    """
    Simulates a ring of the protoplanetary disk

    img_size: int
    The size of the image (height and width)

    alpha: float
    The rotation angle of the core

    center: tuple of float
    The center of the core

    ratio: float
    The ratio the minor axis should be of the major axis a / b

    m: float
    Empirical value determining the scale of the ring

    n: float
    Empirical value determining the scale of the ring

    initial_a: float, optional
    The initial size of the major axis

    device: str, optional
    The name of the device to run the simulations on (with torch)

    Returns
    -------
    ring: torch.tensor
    The image of the ring

    d_ring: float
    The average distance from the center of the ring

    max_value: float
    Largest extent of the ring

    """

    a = np.random.randint(10, 35) + initial_a
    exp = np.random.randint(5, 10)

    ring_a, d_ring_a = _create_ellipse(
        img_size, a, a / ratio, center, alpha, device=device
    )

    ring_b, d_ring_b = _create_ellipse(
        img_size, a + exp, (a + exp) / ratio, center, alpha, device=device
    )
    ring, d_ring = _substract(ring_b, ring_a, d_ring_b, d_ring_a)

    s = _get_smooth_factor(a)
    ring = _smooth(s, s, ring, device=device) * _get_scale(d_ring, m, n)

    max_value = np.max([a, a + exp, a / ratio, (a + exp) / ratio])

    return ring, d_ring, max_value + 6 * s


def _get_smooth_factor(a: float):
    """
    Empirically determines the standard deviation for the gaussian smoothing.
    """

    if a < 15:
        return 2
    if a < 20:
        return 4
    if a < 30:
        return 6
    if a < 40:
        return 9
    if a >= 40:
        return 12


# from https://stackoverflow.com/a/47979802
def _convolve2d(x: torch.tensor, y: torch.tensor):
    """
    Convolves two 2-dimensional images with eachother

    Parameters
    ----------
    x: torch.tensor
    The first image

    y: torch.tensor
    The second image

    Returns
    -------
    cc: torch.tensor
    The convolved image

    """

    fr = fft2(x)
    fr2 = fft2(torch.flipud(torch.fliplr(y)))
    m, n = fr.shape
    cc = torch.real(ifft2(fr * fr2))
    cc = torch.roll(cc, -int(m / 2) + 1, dims=0)
    cc = torch.roll(cc, -int(n / 2) + 1, dims=1)
    return cc


def _smooth(x: float, y: float, c: torch.tensor, device: str = "cpu"):
    """
    Smooths the input image by convolving it with a gaussian 2d-kernel

    Parameters
    ----------
    x: float
    The standard deviation in x-direction

    y: float
    The standard deviation in y-direction

    c: torch.tensor
    The image to smooth

    device: str, optional
    The name of the device to run the simulations on (with torch)

    """

    gauss_kernel = torch.from_numpy(
        Gaussian2DKernel(
            x_stddev=x, y_stddev=y, x_size=c.shape[0], y_size=c.shape[1]
        ).array
    ).to(device)
    return _convolve2d(c, gauss_kernel)


def _e(x: float, m: float, n: float):
    """
    Computes the value of an exponential function exp(m * x + n)
    """

    return np.exp(m * x + n)


def _get_m_n():
    """
    Get the empirical values for m and n from JSON files
    """

    results = {
        "average_1": {"m": -0.08908080536281453, "n": -4.808388735775005},
        "average_2": {"m": -0.04053517757866287, "n": -5.718461569072661},
        "average_3": {"m": -0.023989028595164514, "n": -6.22105762345096},
        "average_4": {"m": -0.011198837787046627, "n": -6.701733964123176},
    }

    id = np.random.choice(list(results.keys()))
    return results[id]["m"], results[id]["n"]


def _get_scale(distance: float, m: float, n: float, noise_scale: float = 0.05):
    """
    Get the empirical standard deviation for the smoothing
    """

    return _e(distance, m, n) + np.random.normal(0, _e(distance, m, n) * noise_scale)



from scipy import stats

__all__ = ["draw_from_mojave_dist", "get_start_amp"]


def get_start_amp(scale_type):
    if scale_type == "mojave":
        amp_start = draw_from_mojave_dist()
    return amp_start / 1e3


def draw_from_mojave_dist():
    """Values from a fit to the peak fluxes distribution
    in the MOJAVE data archive.
    """
    a = 0.8639672251677816
    b = 47.64189171625089
    loc = 0.09163404776954732
    scale = 1892.3881692069087
    return stats.beta(a, b, loc, scale).rvs(1)



# this is adapted from https://github.com/astropy/astropy/blob/master/astropy/version.py
# see https://github.com/astropy/astropy/pull/10774 for a discussion on why this needed.
try:
    try:
        from ._dev_version import version
    except ImportError:
        from ._version import version
except Exception:
    import warnings

    warnings.warn(
        "Could not determine radiosim version. This indicates"
        " a broken installation. Please install radiosim from"
        " the local git repository.",
        stacklevel=2,
    )
    del warnings
    version = "0.0.0"

__version__ = version



import numpy as np

from radiosim.gauss import twodgaussian
from radiosim.jet import create_jet

__all__ = ["create_survey"]


def create_survey(grid, conf):
    """
    Creates a clean survey with all its components written in
    a list. It contains serveral classes:
    - jets
    - gaussian
    - pointsources

    Parameters
    ----------
    grid: 4darray
        input grid of shape [n, 3, img_size, img_size]
    conf:
        loaded config file

    Returns
    -------
    survey: ndarray
        image of the survey, sum over all sources,
        shape: [n, 1, img_size, img_size]
    class_survey: ndarray
        images of each class and background class,
        shape: [n, c + 1, img_size, img_size]
        with c being the max number of classes.
    source_lists: ndarray
        array which stores all properties of each component, shape: [n, 1, 6]
        for each image the class label and its properties.
    """
    class_distribution = conf["class_distribution"]
    scale_sources = conf["scale_sources"]

    if len(grid.shape) == 3:
        grid = grid[None]

    img_size = grid.shape[-1]
    survey = np.zeros((grid.shape[0], 1, img_size, img_size))
    survey_comps = np.zeros(
        (grid.shape[0], len(class_distribution) + 1, img_size, img_size)
    )
    # source_list = []
    for i_img, img in enumerate(grid):
        for _ in range(conf["num_sources"]):
            rand_class = np.random.uniform(0, sum(class_distribution))

            # create first class (jet)
            if rand_class < class_distribution[0]:
                if scale_sources:
                    jet_size = np.random.randint(img_size / 10.24, img_size / 5.12)
                else:
                    jet_size = np.random.randint(100, 200)
                x, y = np.random.rand(2) * img_size
                jet, _ = create_jet(img[0:1, 0:jet_size, 0:jet_size], conf)
                posx_min = int(np.floor(x - jet_size / 2))
                posx_max = int(np.floor(x + jet_size / 2))
                posy_min = int(np.floor(y - jet_size / 2))
                posy_max = int(np.floor(y + jet_size / 2))
                jet = np.squeeze(jet)
                if posx_min < 0:
                    jet = jet[-posx_min:jet_size, :]
                    posx_min = 0
                if posy_min < 0:
                    jet = jet[:, -posy_min:jet_size]
                    posy_min = 0
                if posx_max > img_size:
                    jet = jet[0 : img_size - posx_max + jet_size, :]
                if posy_max > img_size:
                    jet = jet[:, 0 : img_size - posy_max + jet_size]
                survey_comps[i_img, 0, posx_min:posx_max, posy_min:posy_max] += jet
                # source_list.append(jet_list)

            # create second class (gaussian)
            elif rand_class < class_distribution[0] + class_distribution[1]:
                if scale_sources:
                    sx = img_size * np.random.uniform(1 / 40.96, 1 / 20.48)
                    sy = sx + np.random.normal(scale=img_size / 204.8)
                else:
                    sx = np.random.uniform(25, 50)
                    sy = sx + np.random.normal(scale=5)
                x, y = np.random.rand(2) * img_size
                amp = np.random.rand()
                rot = np.random.uniform(0, np.pi)
                gauss = twodgaussian([amp, x, y, sx, sy, rot], img_size)
                survey_comps[i_img, 1] += gauss
                # source_list.append([amp, x, y, sx, sy, rot])

            # create third class (pointsources)
            else:
                if scale_sources:
                    sx = img_size * np.random.uniform(1 / 204.8, 1 / 102.4)
                    sy = sx + np.random.normal(scale=img_size / 1024)
                else:
                    sx = np.random.uniform(5, 10)
                    sy = sx + np.random.normal(scale=1)
                x, y = np.random.rand(2) * img_size
                amp = np.random.rand()
                rot = np.random.uniform(0, np.pi)
                gauss = twodgaussian([amp, x, y, sx, sy, rot], img_size)
                survey_comps[i_img, 2] += gauss
                # source_list.append([amp, x, y, sx, sy, rot])

        survey[i_img] = np.sum(survey_comps[i_img], axis=0)

        # normalisation
        survey_max = survey[i_img].max()
        survey[i_img] /= survey_max
        survey_comps[i_img] /= survey_max
        # for source in source_list:
        #     if isinstance(source, list):
        #         source[0] /= survey_max
        #     if isinstance(source, np.ndarray):
        #         source[0, :, 0] /= survey_max

        # '1 - normalised' gives the background strength
        survey_comps[i_img, -1] = 1 - survey[i_img]

    # set source list to 0, because the saving of the list with variable lengths is
    # not supported
    # source_list = 0

    return (
        survey,
        survey_comps,
        # source_lists,
    )



import numpy as np

from radiosim.flux_scaling import get_start_amp
from radiosim.gauss import twodgaussian
from radiosim.utils import pol2cart, relativistic_boosting, zoom_on_source, zoom_out

__all__ = [
    "apply_train_type",
    "component_from_list",
    "create_jet",
]


def create_jet(grid, conf):
    """
    Creates the clean jets with all its components written
    in a list. Dependend on the 'train_type' the components
    will be seperated or summed up.

    Parameters
    ----------
    grid: ndarray
        input grid of shape [n, 1, img_size, img_size] or
        [1, img_size, img_size]
    conf:
        loaded config file

    Returns
    -------
    jets: ndarray
        image of the full jet, sum over all components,
        shape: [n, 1, img_size, img_size]
    jet_comps: ndarray
        images of each component and background,
        shape: [n, c*2, img_size, img_size] with c being the
        max number of components. A jet without counter jet
        has c components. A jet with counter jet has c*2-1
        components, since the center appears only once. Adding
        one channel for the backgound gives c*2 channels.
    source_lists: ndarray
        array which stores all (seven) properties of each component,
        shape: [n, c*2-1, 7]
    """
    num_comps = conf["num_jet_components"]

    if len(grid.shape) == 3:
        grid = grid[None]

    img_size = grid.shape[-1]
    center = img_size // 2
    jets = []
    targets = []
    for _ in grid:
        comps = np.random.randint(num_comps[0], num_comps[1] + 1)

        amp = np.zeros(num_comps[1])
        x = np.zeros(num_comps[1])
        y = np.zeros(num_comps[1])
        sx = np.zeros(num_comps[1])
        sy = np.zeros(num_comps[1])
        rotation = np.zeros(num_comps[1])
        rotation[0] = np.random.uniform(0, np.pi)

        # velocity in units of c_0, initialise velocity of first component
        beta = np.zeros(num_comps[1])
        beta[1] = np.random.uniform(0, 1)
        y_rotation = np.random.uniform(0, np.pi)
        z_rotation = np.random.uniform(0, np.pi / 2)

        for i in range(comps):
            # amplitude decreases for more distant components, empirical
            amp[i] = np.exp(-np.sqrt(i) * np.random.normal(1.3, 0.2))
            if i >= 1 and np.random.rand() < 0.1:  # drop some components
                amp[i] = 0

            # velocity decreases for more distant components, empirical
            if i >= 2:
                beta[i] = beta[1] * np.exp(-np.sqrt(i - 1) * np.random.normal(0.5, 0.1))

            # curving the jet, empirical
            y_rotation += np.random.normal(0, np.pi / 24)

            # distance between components, r_factor to fill the corners
            jet_angle_cos = np.abs(np.cos(y_rotation))
            jet_angle_sin = np.abs(np.sin(y_rotation))
            if jet_angle_cos < jet_angle_sin:
                r_factor = 1 / jet_angle_sin
            elif jet_angle_cos > jet_angle_sin:
                r_factor = 1 / jet_angle_cos
            else:
                r_factor = np.sqrt(2)

            # *0.7 so the last component is not on the edge
            r = i / (comps - 1) * img_size / 2 * r_factor * np.sin(z_rotation) * 0.7

            # get the cartesian coordinates
            x[i], y[i] = np.array(pol2cart(r, y_rotation)) + center

            # width of gaussian, empirical, sx > sy because rotation up
            # to pi 'changes' this property - fixed to have consistency
            sx[i], sy[i] = np.sort(
                img_size
                / comps
                * r_factor
                * np.sqrt(i + 1)
                / np.random.uniform(3, 8, size=2)
            )[::-1]

            # rotation aligned with the jet angle, empirical
            if i >= 1:
                rotation[i] = rotation[i - 1] + np.random.normal(0, np.pi / 18)

        # print('Velocity of the jet:', beta)
        boost_app, boost_rec = relativistic_boosting(z_rotation, beta)

        center_shift_x = np.random.uniform(-img_size / 20, img_size / 20)
        center_shift_y = np.random.uniform(-img_size / 20, img_size / 20)

        if conf["scaling"] == "mojave":
            amp *= get_start_amp("mojave")

        # mirror the data for the counter jet
        # random drop of counter jet, because the relativistic
        # boosting only does not create clear one-sided jets
        if np.random.rand() < 0.3:
            amp = np.concatenate((amp * boost_app, amp[1:] * boost_rec[1:]))
            x = np.concatenate((x + center_shift_x, img_size - x[1:] + center_shift_x))
            y = np.concatenate((y + center_shift_y, img_size - y[1:] + center_shift_y))
            sx = np.concatenate((sx, sx[1:]))
            sy = np.concatenate((sy, sy[1:]))
            rotation = np.concatenate((rotation, rotation[1:]))
            z_rotation = np.repeat(z_rotation, 2 * num_comps[1] - 1)
            beta = np.concatenate((beta, beta[1:]))
        else:
            amp = np.concatenate((amp * boost_app, np.zeros(num_comps[1] - 1)))
            x = np.concatenate((x + center_shift_x, np.zeros(num_comps[1] - 1)))
            y = np.concatenate((y + center_shift_y, np.zeros(num_comps[1] - 1)))
            sx = np.concatenate((sx, np.zeros(num_comps[1] - 1)))
            sy = np.concatenate((sy, np.zeros(num_comps[1] - 1)))
            rotation = np.concatenate((rotation, np.zeros(num_comps[1] - 1)))
            z_rotation = np.repeat(z_rotation, 2 * num_comps[1] - 1)
            beta = np.concatenate((beta, np.zeros(num_comps[1] - 1)))

        # creation of the image
        jet_comp = np.array(component_from_list(img_size, amp, x, y, sx, sy, rotation))
        jet_img = np.sum(jet_comp, axis=0)

        # get values at the edge of the image
        edge_list = [
            jet_img[0, :-1],
            jet_img[:-1, -1],
            jet_img[-1, ::-1],
            jet_img[-2:0:-1, 0],
        ]
        edges = np.concatenate(edge_list)
        edge_threshold = 0.01
        if edges.max() < edge_threshold:
            # zoom on source to equalize size differences from z-rotation
            jet_img, jet_comp, zoom_factor = zoom_on_source(
                jet_img, jet_comp, max_amp=edge_threshold
            )
            x = img_size / 2 + (x - img_size / 2) * zoom_factor
            y = img_size / 2 + (y - img_size / 2) * zoom_factor
            sx *= zoom_factor
            sy *= zoom_factor

            # random zoom out for more variance
            zoom_out_factor = np.random.uniform(1 / 2, 1)  # 1/8: pad eg. 128 -> 1024
            pad_value = (1 / zoom_out_factor - 1) * img_size / 2
            jet_img, jet_comp = zoom_out(jet_img, jet_comp, pad_value=pad_value)
            x = img_size / 2 + (x - img_size / 2) * zoom_out_factor
            y = img_size / 2 + (y - img_size / 2) * zoom_out_factor
            sx *= zoom_out_factor
            sy *= zoom_out_factor

        # normalisation
        if conf["scaling"] == "normalize":
            jet_max = jet_img.max()
            jet_img /= jet_max
            jet_comp /= jet_max
            amp /= jet_max

        jets.append(jet_img)

        jet_comp = np.concatenate((jet_comp, (1 - jet_img)[None, :, :]))
        source_list = np.array([amp, x, y, sx, sy, rotation, z_rotation, beta]).T

        target = apply_train_type(conf["training_type"], jet_img, jet_comp, source_list)

        targets.append(target)

    jets = np.array(jets)[:, None, :, :]
    targets = np.array(targets)

    return jets, targets


def apply_train_type(train_type, jet_img, jet_comp, source_list):
    """
    Creating the y-data dependent on the training type.

    Parameters
    ----------
    train_type: str
        'list': returns the components attributes only
        'gauss': returns all components, background and list
        'clean': returns sum of components and background (usage for softmax)
    jet_comps: ndarray
        simulated jet components as an image
    source_list:
        attributes of jet components

    Returns
    -------
    y: ndarray
        output data
    """
    if train_type == "list":
        y = source_list
    if train_type == "gauss":
        size = jet_comp.shape[-1]
        list_to_add = np.empty((1, size, size))
        list_to_add[:] = np.nan
        list_to_add[:, 0 : source_list.shape[0], 0 : source_list.shape[1]] = source_list
        y = np.concatenate((jet_comp, list_to_add))
    if train_type == "clean":
        y = jet_img[None]

    return y


def component_from_list(size, amp, x, y, sx, sy, rotation):
    """
    Creating jet components from a list of attributes.

    Parameters
    ----------
    size: int
        shape of the output image will be (size, size)
    attributes: list or array
        [amp, x, y, sx, sy, rotation]

    Returns
    -------
    jet_comp: list of ndarray
        all jet components stored in one list
    """
    jet_comp = []
    for i in range(len(amp)):
        if amp[i] == 0:
            jet_comp += [np.zeros((size, size))]
        else:
            g = twodgaussian(
                [amp[i], x[i], y[i], sx[i], sy[i], rotation[i]],
                size,
            )
            jet_comp += [g]

    return jet_comp
